#! /usr/bin/env python
# vim: expandtab shiftwidth=4 softtabstop=4 tabstop=17 filetype=python :

# Make things as three-ish as possible (requires python >= 2.6)
from __future__ import (unicode_literals, print_function,
                        absolute_import, division)
# Namespace cleanup
del unicode_literals, print_function, absolute_import, division

#
# ----- End header -----
#

import os
import sys

from sqlalchemy import engine_from_config

from pyramid.paster import (
    get_appsettings,
    setup_logging,
    )

from ..models import init_session


def usage(argv):
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri>\n'
          '(example: "%s development.ini")' % (cmd, cmd))
    sys.exit(1)


def main(argv=sys.argv):
    if len(argv) != 2:
        usage(argv)
    config_uri = argv[1]
    setup_logging(config_uri)
    settings = get_appsettings(config_uri)
    engine = engine_from_config(settings, 'sqlalchemy.')
    init_session(engine, create=True)



#! /usr/bin/env python
# vim: expandtab shiftwidth=4 softtabstop=4 tabstop=17 filetype=python :

# Make things as three-ish as possible (requires python >= 2.6)
from __future__ import (unicode_literals, print_function,
                        absolute_import, division)
# Namespace cleanup
del unicode_literals, print_function, absolute_import, division

#
# ----- End header -----
#

from pyramid.config import Configurator
from sqlalchemy import engine_from_config

from .models import (
    init_session,
    )


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    engine = engine_from_config(settings, "sqlalchemy.")
    init_session(engine)
    config = Configurator(settings=settings)
    config.add_route("ca", "/root.crt", request_method="GET")
    config.add_route("cabundle", "/bundle.crt", request_method="GET")
    config.add_route("csr", "/{sha256:[0-9a-f]{64}}", request_method="POST")
    config.add_route("cert", "/{sha256:[0-9a-f]{64}}", request_method="GET")
    config.scan()
    return config.make_wsgi_app()



#! /usr/bin/env python
# vim: expandtab shiftwidth=4 softtabstop=4 tabstop=17 filetype=python :

# Make things as three-ish as possible (requires python >= 2.6)
from __future__ import (unicode_literals, print_function,
                        absolute_import, division)
# Namespace cleanup
del unicode_literals, print_function, absolute_import, division

#
# ----- End header -----
#

import sqlalchemy as _sa
from sqlalchemy.ext.declarative import (
    declarative_base as _declarative_base,
    declared_attr as _declared_attr,
    )
import sqlalchemy.orm as _orm
from zope.sqlalchemy import register

import OpenSSL.crypto as _crypto
from pyramid.decorator import reify as _reify
import datetime as _datetime
import dateutil.parser
import uuid


X509_V3 = 0x2  # RFC 2459, 4.1.2.1

# Bitlength to Hash Strength lookup table.
HASH = {1024: "sha1",
        2048: "sha256",
        4096: "sha512"}

# These parts of the subject _must_ match our CA key
CA_SUBJ_MATCH = (b"C", b"ST", b"L", b"O")


def _crypto_patch():
    """hijack _crypto internal lib and violate the default text encoding.
    https://github.com/pyca/pyopenssl/pull/115 has a pull&fix for it
    https://github.com/pyca/pyopenssl/issues/129 is an open issue
    about it."""
    _crypto._lib.ASN1_STRING_set_default_mask_asc(b'utf8only')


_crypto_patch()


class SigningCert(object):
    """Data class to wrap signing key + cert, to help refactoring"""
    def __init__(self, cert, key=None):
        if key:
            self.key = _crypto.load_privatekey(_crypto.FILETYPE_PEM, key)
        self.cert = _crypto.load_certificate(_crypto.FILETYPE_PEM, cert)

    @classmethod
    def from_files(cls, certfile, keyfile=None):
        key = None
        if keyfile:
            with open(keyfile, 'rt') as f:
                key = f.read()
        with open(certfile, 'rt') as f:
            cert = f.read()

        return cls(cert, key)

    @_reify
    def not_before(self):
        ts = self.cert.get_notBefore()
        if not ts:
            return None
        return dateutil.parser.parse(ts)

    @_reify
    def pem(self):
        return _crypto.dump_certificate(_crypto.FILETYPE_PEM, self.cert)

    # Returns the parts we _care_ about in the subject, from a ca
    def get_ca_prefix(self, subj_match=CA_SUBJ_MATCH):
        subject = self.cert.get_subject()
        components = dict(subject.get_components())
        matches = tuple((n.decode("utf8"), components[n].decode("utf8"))
                        for n in subj_match
                        if n in components)
        return matches


# XXX: probably error prone for cases where things are specified by string
def _fkcolumn(referent, *args, **kwargs):
    refcol = referent.property.columns[0]
    return _sa.Column(refcol.type, _sa.ForeignKey(referent), *args, **kwargs)


DBSession = _orm.scoped_session(_orm.sessionmaker())
register(DBSession)


class Base(object):
    @_declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

    id = _sa.Column(_sa.Integer, primary_key=True)

    def save(self):
        DBSession.add(self)
        DBSession.flush()

    @classmethod
    def query(cls):
        return DBSession.query(cls)

    @classmethod
    def all(cls):
        return cls.query().all()


# XXX: Newer versions of sqlalchemy have a decorator variant 'as_declarative'
Base = _declarative_base(cls=Base)


# XXX: not the best of names
def init_session(engine, create=False):
    DBSession.configure(bind=engine)
    if create:
        Base.metadata.create_all(engine)
    else:
        Base.metadata.bind = engine


# Upper bounds from RFC 5280
_UB_CN_LEN = 64
_UB_OU_LEN = 64

# Length of hex digest of a sha256 checksum
_SHA256_LEN = 64


class CSR(Base):
    sha256sum = _sa.Column(_sa.CHAR(_SHA256_LEN), unique=True, nullable=False)
    pem = _sa.Column(_sa.LargeBinary, nullable=False)
    orgunit = _sa.Column(_sa.String(_UB_OU_LEN))
    commonname = _sa.Column(_sa.String(_UB_CN_LEN))
    rejected = _sa.Column(_sa.Boolean(create_constraint=True))
    accessed = _orm.relationship("AccessLog", backref="csr",
                                 order_by="AccessLog.when.desc()")
    certificates = _orm.relationship("Certificate", backref="csr",
                                     order_by="Certificate.not_after.desc()",
                                     lazy="dynamic",
                                     cascade="all, delete-orphan")

    def __init__(self, sha256sum, reqtext):
        # XXX: assert sha256(reqtext).hexdigest() == sha256sum ?
        self.sha256sum = sha256sum
        self.pem = reqtext
        # FIXME: Below 4 lines (try/except) are duped in the req() function.
        try:
            self.req.verify(self.req.get_pubkey())
        except _crypto.Error:
            raise ValueError("invalid PEM reqtext")
        # Check for and reject reqtext with trailing content
        pem = _crypto.dump_certificate_request(_crypto.FILETYPE_PEM, self.req)
        if pem != self.pem:
            raise ValueError("invalid PEM reqtext")
        self.orgunit = self.subject.OU
        self.commonname = self.subject.CN
        self.rejected = False

    @_reify
    def req(self):
        req = _crypto.load_certificate_request(_crypto.FILETYPE_PEM, self.pem)
        try:
            req.verify(req.get_pubkey())
        except _crypto.Error:
            raise ValueError("Invalid Request")
        return req

    @_reify
    def subject(self):
        return self.req.get_subject()

    @_reify
    def subject_components(self):
        components = self.subject.get_components()
        return tuple((n.decode("utf8"), v.decode("utf8"))
                     for n, v in components)

    @classmethod
    def valid(cls):
        return cls.query().filter_by(rejected=False).all()

    @classmethod
    def refreshable(cls):
        """Using "valid" and looking at csr.certificates doesn't scale.
        Better to do it in the Query."""

        # Options subqueryload is to prevent thousands of small queries and
        # instead batch load the certificates at once
        all_signed = _sa.select([Certificate.csr_id])
        return cls.query().\
            filter_by(rejected=False).\
            filter(CSR.id.in_(all_signed)).all()

    @classmethod
    def unsigned(cls):
        all_signed = _sa.select([Certificate.csr_id])
        return cls.query().filter_by(rejected=False).\
            filter(CSR.id.notin_(all_signed)).all()

    @classmethod
    def by_sha256sum(cls, sha256sum):
        return cls.query().filter_by(sha256sum=sha256sum).one()

    def __json__(self, request):
        url = request.route_url("cert", sha256=self.sha256sum)
        return dict(sha256=self.sha256sum, url=url)

    def __str__(self):
        return ("<{0.__class__.__name__} "  # auto-concatenation (no comma)
                "sha256sum={0.sha256sum:8.8}... "
                "rejected: {0.rejected!r} "
                "OU={0.orgunit!r} CN={0.commonname!r}>").format(self)

    def __repr__(self):
        return ("<{0.__class__.__name__} id={0.id} "  # (no comma)
                "sha256sum={0.sha256sum}>").format(self)


class AccessLog(Base):
    # XXX: name could be better
    when = _sa.Column(_sa.DateTime, default=_datetime.datetime.utcnow)
    # XXX: name could be better, could perhaps be limited length,
    #      might not want this nullable
    addr = _sa.Column(_sa.Text)
    csr_id = _fkcolumn(CSR.id, nullable=False)

    def __init__(self, csr, addr):
        self.csr = csr
        self.addr = addr

    def __str__(self):
        return ("<{0.__class__.__name__} id={0.id} "
                "csr={0.csr.sha256sum} when={0.when}>").format(self)

    def __repr__(self):
        return "<{0.__class__.__name__} id={0.id}>".format(self)


class Extension(object):
    """Convenience class to make validating Extensions a bit easier"""
    critical = False
    name = None
    data = None
    text = None

    def __init__(self, ext):
        self.name = ext.get_short_name()
        self.critical = bool(ext.get_critical())
        self.data = ext.get_data()
        self.text = str(ext)


class Certificate(Base):
    pem = _sa.Column(_sa.LargeBinary, nullable=False)
    # XXX: not_after might be enough
    not_before = _sa.Column(_sa.DateTime, nullable=False)
    not_after = _sa.Column(_sa.DateTime, nullable=False)
    csr_id = _fkcolumn(CSR.id, nullable=False)

    def __init__(self, CSR, pem,  *args, **kws):
        self.pem = pem
        self.csr_id = CSR.id

        req = CSR.req
        cert_pkey = self.cert.get_pubkey()

        # We can't compare pubkeys directly, so we just verify the signature.
        if not req.verify(cert_pkey):
            raise ValueError("Public key of cert cannot verify request")

        self.not_before = dateutil.parser.parse(self.cert.get_notBefore())
        self.not_after = dateutil.parser.parse(self.cert.get_notAfter())

    @_reify
    def cert(self):
        cert = _crypto.load_certificate(_crypto.FILETYPE_PEM, self.pem)

        extensions = {}
        for index in range(0, cert.get_extension_count()):
            ext = cert.get_extension(index)
            my_ext = Extension(ext)
            extensions[my_ext.name] = my_ext

        if cert.get_version() != X509_V3:
            raise ValueError("Not a x509.v3 certificate")

        ext = extensions.get(b'basicConstraints')
        if not ext:
            raise ValueError("Missing Basic Constraints")

        if not ext.critical:
            raise ValueError("Extended Key Usage not critical")

        ext = extensions.get(b'extendedKeyUsage')
        if not ext:
            raise ValueError("Missing Extended Key Usage extension")
        if not ext.critical:
            raise ValueError("Extended Key Usage not critical")
        if "TLS Web Client Authentication" in ext.text:
            pass
        if "TLS Web Server Authentication" in ext.text:
            pass
        return cert

    def __repr__(self):
        return "<{0.__class__.__name__} id={0.id}>".format(self)

    @classmethod
    def sign(cls, CSR, ca, lifetime=_datetime.timedelta(30*3),
             backdate=False):
        """Takes a CSR, signs it, generating and returning a Certificate.
        backdate causes the CA to set "notBefore" of signed certificates to
        match that of the CA Certificate. This is an ugly workaround for a
        timekeeping bug in some firmware.
        """
        assert isinstance(ca, SigningCert)
        notAfter = int(lifetime.total_seconds())
        # TODO: Verify that the data in DB matches csr_add rules in views.py

        cert = _crypto.X509()
        cert.set_subject(CSR.req.get_subject())
        cert.set_serial_number(int(uuid.uuid1()))
        cert.set_issuer(ca.cert.get_subject())
        cert.set_pubkey(CSR.req.get_pubkey())
        cert.set_version(X509_V3)
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(notAfter)
        if backdate and ca.not_before:
            now = _datetime.datetime.now(tz=_datetime.timezone.utc)
            delta = ca.not_before - now
            cert.gmtime_adj_notBefore(int(delta.total_seconds()))

        subjectAltName = bytes("DNS:" + CSR.commonname, 'utf-8')
        extensions = [
            _crypto.X509Extension(b'basicConstraints',
                                  critical=True,
                                  value=b'CA:FALSE'),
            _crypto.X509Extension(b'extendedKeyUsage',
                                  critical=True,
                                  value=b'clientAuth,serverAuth'),
            _crypto.X509Extension(b"subjectAltName",
                                  critical=False,
                                  value=subjectAltName),
            _crypto.X509Extension(b"subjectKeyIdentifier",
                                  critical=False,
                                  value=b"hash",
                                  subject=cert),
        ]
        cert.add_extensions(extensions)
        # subjectKeyIdentifier has to be present before adding auth ident
        extensions = [
            _crypto.X509Extension(b"authorityKeyIdentifier",
                                  critical=False,
                                  value=b"issuer:always,keyid:always",
                                  issuer=ca.cert)
        ]
        cert.add_extensions(extensions)
        bits = cert.get_pubkey().bits()
        cert.sign(ca.key, HASH[bits])
        pem = _crypto.dump_certificate(_crypto.FILETYPE_PEM, cert)
        return cls(CSR=CSR, pem=pem)



#! /usr/bin/env python
# vim: expandtab shiftwidth=4 softtabstop=4 tabstop=17 filetype=python :

# Make things as three-ish as possible (requires python >= 2.6)
from __future__ import (unicode_literals, print_function,
                        absolute_import, division)
# Namespace cleanup
del unicode_literals, print_function, absolute_import, division

#
# ----- End header -----
#

from textwrap import dedent
from hashlib import sha256
from operator import attrgetter

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

from datetime import datetime, timedelta
day = timedelta(days=1)
year = 365 * day                # close enough
now = datetime.now()

from caramel import (
    models,
    views,
    )


class defaultproperty(object):
    def __init__(self, func):
        self.__doc__ = getattr(func, "__doc__", None)
        self.func = func

    def __get__(self, obj, objtype=None):
        return self if obj is None else self.func(obj)


class SimilarityComparable(object):
    _match_excluded_attrs_ = ("match",)

    def match(self, other):
        excludes = set(getattr(self, "_match_excluded_attrs_", ()))
        attrs = [attr for attr in dir(self)
                 if not attr.startswith("_") and attr not in excludes]
        if not attrs:
            return True
        getter = attrgetter(*attrs)
        try:
            return getter(self) == getter(other)
        except AttributeError:
            return False


class AttributeCollection(object):
    def __init__(self, _base_=None, **kwds):
        if _base_:
            kwds = dict(_base_._attrs, **kwds)
        for k, v in kwds.items():
            setattr(self, k, v)
        self._attrs = kwds


class CSRFixture(AttributeCollection, SimilarityComparable):
    __relations = ("accessed", "certificates",)

    @property
    def _match_excluded_attrs_(self):
        return (super(CSRFixture, self)._match_excluded_attrs_ +
                self.__relations)

    @defaultproperty
    def sha256sum(self):
        return sha256(self.pem).hexdigest()

    def match(self, other):
        if not super(CSRFixture, self).match(other):
            return False
        for relation in self.__relations:
            related = getattr(self, relation, ())
            if related:
                other_related = getattr(other, relation, ())
                if not other_related:
                    return False
                for a, b in zip_longest(related, other_related):
                    if not a.match(b):
                        return False
        return True

    def __call__(self):
        csr = models.CSR(self.sha256sum, self.pem)
        for relation in self.__relations:
            related = getattr(self, relation, ())
            for relatee in related:
                relatee(csr).save()
        return csr


class CertificateFixture(AttributeCollection, SimilarityComparable):
    def __call__(self, csr):
        # FIXME: adjust when models.Certificate gets a proper __init__
        cert = models.Certificate(csr, self.pem)
        for name, value in self._attrs.items():
            setattr(cert, name, value)
        cert.csr = csr
        return cert


class AccessLogFixture(AttributeCollection, SimilarityComparable):
    def __call__(self, csr):
        access = models.AccessLog(csr, self.addr)
        access.when = self.when
        return access


# "Correct" subject prefix for test data
subject_prefix = (('O', 'Example inc.'), ('OU', 'Example Dept'),)


class CertificateData(object):
    initial = CertificateFixture(
        not_before=now - 2 * year,
        not_after=now + 2 * year,
        pem=dedent("""\
            -----BEGIN CERTIFICATE-----
            MIIEVDCCAjygAwIBAgIBZTANBgkqhkiG9w0BAQsFADAmMSQwIgYDVQQDDBtDYXJh
            bWVsIFNpZ25pbmcgQ2VydGlmaWNhdGUwHhcNMTQwODA4MTM0NDA0WhcNMTQxMDA4
            MTM0NDA0WjAaMRgwFgYDVQQDDA9mb28uZXhhbXBsZS5jb20wggEiMA0GCSqGSIb3
            DQEBAQUAA4IBDwAwggEKAoIBAQC+UOYhTExyWQL/mPzxvkd4xzju9Gcoxu5WZxxP
            +FjvsCKXkdMdzyTGYuaOulsRqyNmc8N73KJa9rlajIxdtR7duBwzX6Ddx2MmOLVL
            Q95X0jzdYR9iv6NX+bdYnoFUuzFt6N6Tf6OFS7IGiCeYqN7JTzTwlseJ7O4ozdsk
            vNlw7Cybb5RjqJQ/TRDSDWp1Fuq7FXanM+9Eaok0xqGty1TdMiEsCK8t7w3F1gd3
            bt48hld7cfe+OqdPxFLLtXv6wVM8EzcFYmwBGx7avXVS8aOsN6Oc3FZBfW0Fi0XD
            S6YO7WSzsRqciXolv/zPIvc1g6z5RkIPlgUaCtAdNBWkDG0hAgMBAAGjgZgwgZUw
            DAYDVR0TAQH/BAIwADAWBgNVHSUBAf8EDDAKBggrBgEFBQcDAjAdBgNVHQ4EFgQU
            79OVse0vdmb3VW6T9P7haHDOrhEwTgYDVR0jBEcwRYAUt9UtYpijdgeh1fObZHE4
            EBJPROGhKqQoMCYxJDAiBgNVBAMMG0NhcmFtZWwgU2lnbmluZyBDZXJ0aWZpY2F0
            ZYIBADANBgkqhkiG9w0BAQsFAAOCAgEAN8oppVaZlklLmvD1OfG0jf7SwWa03x6L
            Ej7xh0EbCeTZXNCmJNjiPpYG57Lf7OlXit9u1aJUQ5VXpUDmv1njb0jnB3EJBWf/
            Y2KZbRYgXCpMUXxmfLrntULHAn/M3RLNi1ovwisiBqAU5/SZLcRq26Uwb8ygT0K7
            OA80JqjVzv53nKdUmDR0mopiWiU8c5BTi0R5+Q9vGov+DBcmrgatIHrtiXG8exGK
            m/DHZMAA93Vv/nh6GUe9uWI4mDhhLRSrdfkHoTPDqrFHIdH4OC+ljlPDRNhpkWAt
            FIzejEk5pqyqnAW4HHWM13vtOpddnXea+1y6PCX/9InAP5Tl1HMcq7BOgQa+EyPN
            NdzIRtdRnhtncMJUTCJpm2QEH09aPM3411tFpG3nZ6eTvo5f6oZS+epBVwNVqFXm
            xTbIEnBFnVBhQh2mXgJJsac2Oy8KGyBnVvGn6FKB6slrBjcIWKg+CDyiXMjwKhk5
            bjUvUdML4uWoGqdJfl+S+f+8m6S6F276p/uLI8Kb0NaE+z+LFRp2wrrC7DW65R+W
            eMdFWSu/IZZkoEaTrOeaqmLtQQt/4s66KKNOddftt+uLEMjce7foYQHo+qx9RI/1
            n4rH2rFn8rcQwSaRyE9NcOpirCur43MR42+LAfZq5s9j7CuQJTw6G0wvCHGqRIQ9
            X1qbQOxg6ig=
            -----END CERTIFICATE-----
            """).encode("utf8"),
        )

    expired = CertificateFixture(
        not_before=initial.not_before,
        not_after=now - 1 * day,
        pem=dedent("""\
            -----BEGIN CERTIFICATE-----
            MIIEVTCCAj2gAwIBAgIBaDANBgkqhkiG9w0BAQsFADAmMSQwIgYDVQQDDBtDYXJh
            bWVsIFNpZ25pbmcgQ2VydGlmaWNhdGUwHhcNMTQwODA4MTM1OTUxWhcNMTQxMDA4
            MTM1OTUxWjAbMRkwFwYDVQQDDBBzcGFtLmV4YW1wbGUuY29tMIIBIjANBgkqhkiG
            9w0BAQEFAAOCAQ8AMIIBCgKCAQEAnrzJ2qNhyIiCZvikFvPr0iEfzf27kTfHN+13
            7bGuTAzmXFFrajz9K+leYcZR4sCPPRI8QjEUSYcngP1LKUrgzUSldjPrQGLsmx+8
            gLdi2JN1kPu6uMT97uB1RDKQpIMHGuV4mJKJku3sh6DJvQMjuMv8xOUXHtCw9jjc
            6CLI9zeEZfM1RXsmRVxLx8HuwlF8ZNRjPGn5lEGIxTORpF1Mef5eTnGIg2kxwj8F
            5aJP4ei9XS6gbYSFZekruT9Ivh41x4FYmx4bcGQfuIu+cG+XI3xGoNrm/IKMguMQ
            vwjBj2XnLHzskR5mz9YSLd9vi1nHl7b5u9GDtpmliNOvehaxjQIDAQABo4GYMIGV
            MAwGA1UdEwEB/wQCMAAwFgYDVR0lAQH/BAwwCgYIKwYBBQUHAwIwHQYDVR0OBBYE
            FH/tkDyXntmyru3UFyYTO6UTOwjuME4GA1UdIwRHMEWAFLfVLWKYo3YHodXzm2Rx
            OBAST0ThoSqkKDAmMSQwIgYDVQQDDBtDYXJhbWVsIFNpZ25pbmcgQ2VydGlmaWNh
            dGWCAQAwDQYJKoZIhvcNAQELBQADggIBAD4r7SEyD6A61ORVjHHpw+5nZuaIf3Xs
            nZakE5I+wm/ZsVmWO3iOjAKdL2HartlDqGAvX5/w/sEoOkrpr48R7Uolm4ojYD2B
            Etxb4p9PDHmgT9Sc8nE2hOh8FdIUiGVLRJOuCmZqgk+EBIwB0zJQjREXHVZaN4dS
            wNNQGz46Vd2b6e4iRCJdgfOd2tuJi/WtvQORJkJ+HeTwLcN9LYY6d1bgups970DP
            Ve3QzxowNaRdX4SNtTmWs2BqcpYJV/Bx+q5r5CxhuDs5Jyvfmh9K7ux8z8Rv9THm
            UPJSXahBKLIYI8BYotXB9vpHvgWd15Opa3g+MV1Ego7KDKvdirTKnIcvafAi47vo
            kpAussllmvGr3wjDrGLlvbyBbbg44MpTzZrWspAjm9t0XFwkDDZGjY9l1zzqolDa
            KXbDda3LApBjWOr2QxsiwuOcemeekyiYusx5O4NqWJhGRVluvLpo5FWvuGZfoTCG
            X9QYUVrbsvhMMZR0VQDkJ7zfqbA8HvZ72qRpkC3MkOKDo4Ve9PffRjK+bamZM38I
            rvKws1E0w1WI1cK8ypefZhF8BiC/KUFMBw3bQAUQqdubv4eN81IX0PjnpfH+7wwA
            7CRL6mRM5+q4ICfesXzUXpPYjgD2UexIjSJNoYjiu3aKpjwvrZ88t8hPR46Uac7E
            5yvm61Agw/Tg
            -----END CERTIFICATE-----
            """).encode("utf8"),
        )

    ca_cert = CertificateFixture(
        not_before=initial.not_before,
        not_after=initial.not_after,
        subject=(('C', 'SE'), ('ST', 'Östergötland'),
                 ('L', 'Norrköping'), ('O', 'Muppar AB'),
                 ('OU', 'Muppar Teknik'),
                 ('CN', 'Caramel Signing Certificate')),

        common_subject=(('C', 'SE'), ('ST', 'Östergötland'),
                        ('L', 'Norrköping'), ('O', 'Muppar AB')),

        pem=dedent("""\
            -----BEGIN CERTIFICATE-----
            MIIGwDCCBKigAwIBAgIRAJSEOECNQRHkq2SMiaXBGsIwDQYJKoZIhvcNAQENBQAw
            gY4xCzAJBgNVBAYTAlNFMRcwFQYDVQQIDA7DlnN0ZXJnw7Z0bGFuZDEUMBIGA1UE
            BwwLTm9ycmvDtnBpbmcxEjAQBgNVBAoMCU11cHBhciBBQjEWMBQGA1UECwwNTXVw
            cGFyIFRla25pazEkMCIGA1UEAwwbQ2FyYW1lbCBTaWduaW5nIENlcnRpZmljYXRl
            MB4XDTE0MTIyNjIwNTU1M1oXDTM4MTIyNjIwNTU1M1owgY4xCzAJBgNVBAYTAlNF
            MRcwFQYDVQQIDA7DlnN0ZXJnw7Z0bGFuZDEUMBIGA1UEBwwLTm9ycmvDtnBpbmcx
            EjAQBgNVBAoMCU11cHBhciBBQjEWMBQGA1UECwwNTXVwcGFyIFRla25pazEkMCIG
            A1UEAwwbQ2FyYW1lbCBTaWduaW5nIENlcnRpZmljYXRlMIICIjANBgkqhkiG9w0B
            AQEFAAOCAg8AMIICCgKCAgEAtgt5ghocut1Qc7voPnpsuSWFvAC08e6LOk8ilQi8
            d9i2he19SFlNCFblhOLNsit+bBwFjWTIMsAaz8e1X7mQCR1sKKr1ShyvSTVw0UdQ
            uRNyYfP83h0vub4gOlJSvN4R816UP6zETHxRtdEZg7hlXY7NJsHnszEkxBicku2M
            an+CvB3//0EBSd20jUcDLR0K4fLNpku1gLrqfnejnZAnvUrW1LI2Nhhi10akiLNT
            a88rWicCwZtxCvwygjLJYL7FfL1NnmZwpYASorqldHu6RoUst3WL1r/dy0mpzftB
            kUG2Docf/3TxVkcdoC/Mjh3NqMAggJ3EdzTDEMGDUUG2EdtXXmT3GMEMW3y49Uww
            i4p36tIzVPpfjEVHFl4SJvafEiyCcCc9BxmtWLvXRFr2Q4dUmyDhdtNj9kT1w/94
            V71msSVlWdNvm7vPlV75tNtNAzdmFgs5Rnakkkc4QKyOttPJ+Cl+SO07fhziNVSU
            5AWXLEHwJ9oUMp7H/FrvPwsQhMQIQ1dzQVoBTH473v3GRXFTwLseGgzITIN+vVv6
            K0ap1H/SkLfwuHMev6xXUF43w8kaDSvjHTN9Jad2IdjQ8siBuLlRBiLU4REn60lQ
            E/v5ttj2MWjXBIbt9yMWSYXtQCZTya778PZrGUJwIfEyzHfO2O0nODLuvuafxveo
            ShUCAwEAAaOCARUwggERMBIGA1UdEwEB/wQIMAYBAf8CAQAwDgYDVR0PAQH/BAQD
            AgIEMB0GA1UdDgQWBBQET+HX8lWrMux6W0Nd5heTB0TKWjCBywYDVR0jBIHDMIHA
            gBQET+HX8lWrMux6W0Nd5heTB0TKWqGBlKSBkTCBjjELMAkGA1UEBhMCU0UxFzAV
            BgNVBAgMDsOWc3RlcmfDtnRsYW5kMRQwEgYDVQQHDAtOb3Jya8O2cGluZzESMBAG
            A1UECgwJTXVwcGFyIEFCMRYwFAYDVQQLDA1NdXBwYXIgVGVrbmlrMSQwIgYDVQQD
            DBtDYXJhbWVsIFNpZ25pbmcgQ2VydGlmaWNhdGWCEQCUhDhAjUER5KtkjImlwRrC
            MA0GCSqGSIb3DQEBDQUAA4ICAQAx3jr30fHu9/AJpCnzs1nzNDoGwN7YdyZIWPUm
            uZJ2TPpMukK1bI38LXYS/bmjhgmc3MoyEridImm+FtFWFqDNASwcYpPEucH1HgEv
            gVp8MznhZIEJhB799hrDqLh3HRKAbgV3bz4zPFr3V6R3YRrMekVCBtcFHdDKq1Uy
            b4HwalrK6ZVxSr92DJy0Qyk5Zhtz1RGK1+7uECqWq5K5pTSMcw+MiLlESX+Brz4D
            EKABp4cLsOpHglasSNUVrmOaDg2qbz96PhsergjOeYQbakK87H/98pusFyPHkmYR
            nPbwiJi2D/9kxKwNkLiNI5oWRWY1ldcEE7C8hi8tOegNMu68UA2sx7Nhip3qUVcx
            M2JLOhBiq0BDCb+QpsF0NSeMyJCbkRogdYsIzGmCF51ubq/zYUE5mwvypD7CcdbI
            947sucl+E8bLwcGMvOBEYUdSAisya8Qd7D2h/UwGHlBZu1SX3LCI7GopJ3LvkHG/
            rF6tBrFEeXx1536rChzNal58wY+0HamTJQYFvh88OpbbwLP+UlPm4Gh0Rx3i61/I
            Op+poNAWLT5NACOX12yVfAcz3fWwxCoY62YoSfdUjH10BwZZssZdVIWdOGsCEOaL
            EybN1zenahQmhX4ljI5OSMYBP3gpkRu2HvP5SOwN28QHq2+mgCFxzdfdtiWoJ8a0
            7eLmmg==
            -----END CERTIFICATE-----
            """).encode("utf8"),
        )
    # FIXME: add more certificates here, once we have something to test.


class AccessLogData(object):
    initial_1 = AccessLogFixture(
        when=CertificateData.initial.not_before + 30 * day,
        addr="127.0.0.1",       # XXX: bytestring or unicode string?
        )

    initial_2 = AccessLogFixture(
        when=now - 3 * day,
        addr="127.0.0.127",
        )


class CSRData(object):
    initial = CSRFixture(
        orgunit="Example Dept",
        commonname="foo.example.com",
        pem=dedent("""\
            -----BEGIN CERTIFICATE REQUEST-----
            MIICjTCCAXUCAQIwSDEVMBMGA1UECgwMRXhhbXBsZSBpbmMuMRUwEwYDVQQLDAxF
            eGFtcGxlIERlcHQxGDAWBgNVBAMMD2Zvby5leGFtcGxlLmNvbTCCASIwDQYJKoZI
            hvcNAQEBBQADggEPADCCAQoCggEBAL5Q5iFMTHJZAv+Y/PG+R3jHOO70ZyjG7lZn
            HE/4WO+wIpeR0x3PJMZi5o66WxGrI2Zzw3vcolr2uVqMjF21Ht24HDNfoN3HYyY4
            tUtD3lfSPN1hH2K/o1f5t1iegVS7MW3o3pN/o4VLsgaIJ5io3slPNPCWx4ns7ijN
            2yS82XDsLJtvlGOolD9NENINanUW6rsVdqcz70RqiTTGoa3LVN0yISwIry3vDcXW
            B3du3jyGV3tx9746p0/EUsu1e/rBUzwTNwVibAEbHtq9dVLxo6w3o5zcVkF9bQWL
            RcNLpg7tZLOxGpyJeiW//M8i9zWDrPlGQg+WBRoK0B00FaQMbSECAwEAAaAAMA0G
            CSqGSIb3DQEBCwUAA4IBAQBxkfl/Nq0y2u8Deq17OlB8WZfJnigEtFtMFstWNxNp
            Z3yxSFbaOYpB/+S0qOTjFbx1vQd8sSKTEqSgn2MkLhNsqtakWhejC+rrzVA02K6d
            J7uCylX8XVRJPjmt14E2LNLxGx1adV8St0tPrbzXMzr0ygpGaIITvd+ZzXr4CuGQ
            vao+T3EooEVQeFmWaoU6URUsqlp1itYzN1O+tvHv9pmCZ5UyTJlHQSc+cKJsUgE4
            O+jruZshnFL0KQybIokYGLLcb6NixdsCTSw+rfztuLUEEMP1ozNCgk8TX8mXWduM
            XIP49FHFe6IjLuj0ofRXiJPmS+4ToqRbNIBRoz7kSLov
            -----END CERTIFICATE REQUEST-----
            """).encode("utf8"),  # py3 dedent can't handle bytes
        subject_components=(
            ('O', 'Example inc.'),
            ('OU', 'Example Dept'),
            ('CN', 'foo.example.com'),
        ),
        accessed=[
            AccessLogData.initial_2,
            AccessLogData.initial_1,
        ],
        certificates=[
            CertificateData.initial,
        ],
        )

    with_expired_cert = CSRFixture(
        orgunit="Example Dept",
        commonname="spam.example.com",
        pem=dedent("""\
            -----BEGIN CERTIFICATE REQUEST-----
            MIICjjCCAXYCAQIwSTEVMBMGA1UECgwMRXhhbXBsZSBpbmMuMRUwEwYDVQQLDAxF
            eGFtcGxlIERlcHQxGTAXBgNVBAMMEHNwYW0uZXhhbXBsZS5jb20wggEiMA0GCSqG
            SIb3DQEBAQUAA4IBDwAwggEKAoIBAQCevMnao2HIiIJm+KQW8+vSIR/N/buRN8c3
            7Xftsa5MDOZcUWtqPP0r6V5hxlHiwI89EjxCMRRJhyeA/UspSuDNRKV2M+tAYuyb
            H7yAt2LYk3WQ+7q4xP3u4HVEMpCkgwca5XiYkomS7eyHoMm9AyO4y/zE5Rce0LD2
            ONzoIsj3N4Rl8zVFeyZFXEvHwe7CUXxk1GM8afmUQYjFM5GkXUx5/l5OcYiDaTHC
            PwXlok/h6L1dLqBthIVl6Su5P0i+HjXHgVibHhtwZB+4i75wb5cjfEag2ub8goyC
            4xC/CMGPZecsfOyRHmbP1hIt32+LWceXtvm70YO2maWI0696FrGNAgMBAAGgADAN
            BgkqhkiG9w0BAQsFAAOCAQEAf5SnEDewYNy4bu1bwga2alawbr+BQpysl/h53/aJ
            woJakG3E3+jSyAQ6Bu/j+YY+hTMbNX/sOyWMexS6w6HryQ6i/xvFZqhgN5ap/I6M
            k7V13j1LyxcTtlD773ikWq7F/+H76FgTAoE4oUmNvptej/C5eW5A1N150vMeecY/
            0nHIV+eXaZTCbg7UKr1xmR3tdu3DNnFC/BfaVv+Ul7s974k4g53ejLCvnCjMGVS3
            oQN1HuxKCScUJ9Vtr0dnBLGAf62vAqv5yZYhl9Qnt5EJ9OtspWm0e8FwTjNmoA/z
            qnvUxskzM2ItxVV9oa9YDTid0GbJvF67QJQyVIO0Vz4uwg==
            -----END CERTIFICATE REQUEST-----
            """).encode("utf8"),  # py3 dedent can't handle bytes
        subject_components=(
            ('O', 'Example inc.'),
            ('OU', 'Example Dept'),
            ('CN', 'spam.example.com'),
        ),
        certificates=[
            CertificateData.expired,
        ],
        )

    good = CSRFixture(
        orgunit="Example Dept",
        commonname="bar.example.com",
        pem=dedent("""\
            -----BEGIN CERTIFICATE REQUEST-----
            MIICjTCCAXUCAQIwSDEVMBMGA1UECgwMRXhhbXBsZSBpbmMuMRUwEwYDVQQLDAxF
            eGFtcGxlIERlcHQxGDAWBgNVBAMMD2Jhci5leGFtcGxlLmNvbTCCASIwDQYJKoZI
            hvcNAQEBBQADggEPADCCAQoCggEBAN7fng4vFo0P0+K1L64rADgXBDrwsa39p3tV
            7GwY/LZ9crxUgwFKfVLM8rX4KAySiOXix8JF44jansXTOkcm8OjnOKVNIJX/5Pf3
            bRDSXcjodFIhPVzUynj8E5Z8rEB2ES9gwYDKIYVNnJa2nGmQVe7IgA5O7lNM6gse
            TqYlN3bmB9Dy/dY+ZVyts1p6aSOYMdAcJ7ojCco9HuYFav2hd7k2h4b4lCKvi7p2
            sx6DoKmnblmlvMlP9UdSq7wORkAUnMn5rlzdj5LnWsSB+JLBrQaSlsIfVeL0+5oB
            6zrZI5hJPQiaLcet0v6a7M7UkOyJKiGJVCKYnO122D4Cu4lB2d0CAwEAAaAAMA0G
            CSqGSIb3DQEBCwUAA4IBAQAGJwl7vUlDg2L3KgnsNA9A4rMKfFn5fzdf6Z0X3HSY
            Zvi8XKkVAKd7aRUSg6jcJbOm0HNmBR+5SWzXUU62KQWuUCwz+dCyCbcEO/frG0IB
            HdOs/L4AJgHlIaxwisCLh5VQpj5w0ahhzVfLGWcCK1nbqjUTLcEFhvZviUPUugAD
            f7QdWNDBuZTrwXGTLsnhC5XQfsk0maXcNji79ziK4sMm5TU2599JmuWL2NTbqaCQ
            MN1FxWzEy4yXgzW8uv+lX6yyTtkfrC7e3LFiAuoUlBeD5GmsVd30Xz5iGnuQv3d0
            fekjT5Np8XIS2ERJmx4CIjs5VpE1FMNOMoJ35kQpkQaQ
            -----END CERTIFICATE REQUEST-----
            """).encode("utf8"),
        subject_components=(
            ('O', 'Example inc.'),
            ('OU', 'Example Dept'),
            ('CN', 'bar.example.com'),
        ),
        )

    bad_signature = CSRFixture(  # `good` with the subject of `initial`
        pem=dedent("""\
            -----BEGIN CERTIFICATE REQUEST-----
            MIIBAjCBrQIBADBIMRUwEwYDVQQKDAxFeGFtcGxlIGluYy4xFTATBgNVBAsMDEV4
            YW1wbGUgRGVwdDEYMBYGA1UEAwwPZm9vLmV4YW1wbGUuY29tMFwwDQYJKoZIhvcN
            AQEBBQADSwAwSAJBAKk2sD6xi/gfO3TVnoGMhUmkPDD17/qYzEvDdw/kponLTdNF
            asGx1//giKSBqBpUFt+KTz3NofK9Pf2qWWDxyUECAwEAAaAAMA0GCSqGSIb3DQEB
            BQUAA0EAcsrzTdYBqlbq/JQaMSEoi64NmoxiC8GGzOaKlTxqRc7PKb+T1wN94PxJ
            faXw8kA8p0E6hmwFAE9QVkuTKvP/eg==
            -----END CERTIFICATE REQUEST-----
            """).encode("utf8"),
        )

    truncated = CSRFixture(
        pem=dedent("""\
            -----BEGIN CERTIFICATE REQUEST-----
            MIIBAjCBrQIBADBIMRUwEwYDVQQKDAxFeGFtcGxlIGluYy4xFTATBgNVBAsMDEV4
            YW1wbGUgRGVwdDEYMBYGA1UEAwwPYmFyLmV4YW1wbGUuY29tMFwwDQYJKoZIhvcN
            AQEBBQADSwAwSAJBAKk2sD6xi/gfO3TVnoGMhUmkPDD17/qYzEvDdw/kponLTdNF
            asGx1//giKSBqBpUFt+KTz3NofK9Pf2qWWDxyUECAwEAAaAAMA0GCSqGSIb3DQEB
            """).encode("utf8"),
        )

    trailing_content = CSRFixture(
        pem=dedent("""\
            -----BEGIN CERTIFICATE REQUEST-----
            MIIBAjCBrQIBADBIMRUwEwYDVQQKDAxFeGFtcGxlIGluYy4xFTATBgNVBAsMDEV4
            YW1wbGUgRGVwdDEYMBYGA1UEAwwPYmFyLmV4YW1wbGUuY29tMFwwDQYJKoZIhvcN
            AQEBBQADSwAwSAJBAKk2sD6xi/gfO3TVnoGMhUmkPDD17/qYzEvDdw/kponLTdNF
            asGx1//giKSBqBpUFt+KTz3NofK9Pf2qWWDxyUECAwEAAaAAMA0GCSqGSIb3DQEB
            BQUAA0EAcsrzTdYBqlbq/JQaMSEoi64NmoxiC8GGzOaKlTxqRc7PKb+T1wN94PxJ
            faXw8kA8p0E6hmwFAE9QVkuTKvP/eg==
            -----END CERTIFICATE REQUEST-----
            foo
            bar
            baz
            quux
            """).encode("utf8"),
        )

    leading_content = CSRFixture(
        pem=dedent("""\
            foo
            bar
            baz
            quux
            -----BEGIN CERTIFICATE REQUEST-----
            MIIBAjCBrQIBADBIMRUwEwYDVQQKDAxFeGFtcGxlIGluYy4xFTATBgNVBAsMDEV4
            YW1wbGUgRGVwdDEYMBYGA1UEAwwPYmFyLmV4YW1wbGUuY29tMFwwDQYJKoZIhvcN
            AQEBBQADSwAwSAJBAKk2sD6xi/gfO3TVnoGMhUmkPDD17/qYzEvDdw/kponLTdNF
            asGx1//giKSBqBpUFt+KTz3NofK9Pf2qWWDxyUECAwEAAaAAMA0GCSqGSIb3DQEB
            BQUAA0EAcsrzTdYBqlbq/JQaMSEoi64NmoxiC8GGzOaKlTxqRc7PKb+T1wN94PxJ
            faXw8kA8p0E6hmwFAE9QVkuTKvP/eg==
            -----END CERTIFICATE REQUEST-----
            """).encode("utf8"),
        )

    multi_request = CSRFixture(
        pem=dedent("""\
            -----BEGIN CERTIFICATE REQUEST-----
            MIIBAjCBrQIBADBIMRUwEwYDVQQKDAxFeGFtcGxlIGluYy4xFTATBgNVBAsMDEV4
            YW1wbGUgRGVwdDEYMBYGA1UEAwwPZm9vLmV4YW1wbGUuY29tMFwwDQYJKoZIhvcN
            AQEBBQADSwAwSAJBANBBe43zbCwPs1jRPPqgb6Otdqx9kg67Dgfoxh3Mly7b9JPp
            orKIs6zzoyMeLHLJYk+CwSHyeS1hc4zL+3A+k88CAwEAAaAAMA0GCSqGSIb3DQEB
            BQUAA0EALt6dIjlzG05KCfyiy2PJdAwcjC+mpHh3i4cJs50U+EnxBgX8QscOu382
            72uukmhYBG1Xd1LN4S5RL9pcQ9KLaA==
            -----END CERTIFICATE REQUEST-----
            -----BEGIN CERTIFICATE REQUEST-----
            MIIBAjCBrQIBADBIMRUwEwYDVQQKDAxFeGFtcGxlIGluYy4xFTATBgNVBAsMDEV4
            YW1wbGUgRGVwdDEYMBYGA1UEAwwPYmFyLmV4YW1wbGUuY29tMFwwDQYJKoZIhvcN
            AQEBBQADSwAwSAJBAKk2sD6xi/gfO3TVnoGMhUmkPDD17/qYzEvDdw/kponLTdNF
            asGx1//giKSBqBpUFt+KTz3NofK9Pf2qWWDxyUECAwEAAaAAMA0GCSqGSIb3DQEB
            BQUAA0EAcsrzTdYBqlbq/JQaMSEoi64NmoxiC8GGzOaKlTxqRc7PKb+T1wN94PxJ
            faXw8kA8p0E6hmwFAE9QVkuTKvP/eg==
            -----END CERTIFICATE REQUEST-----
            """).encode("utf8"),
        )

    not_pem = CSRFixture(
        pem=dedent("""\
            Egg and Bacon
            Egg, Sausage and Bacon
            Egg and Spam
            Egg, Bacon and Spam
            Egg, Bacon, Sausage and Spam
            Spam, Bacon, Sausage and Spam
            Spam, Egg, Spam, Spam, Bacon and Spam
            Spam, Spam, Spam, Egg and Spam
            Spam, Spam, Spam, Spam, Spam, Spam, Baked Beans,
                Spam, Spam, Spam and Spam
            """).encode("utf8"),
        )

    empty = CSRFixture(pem=b"")

    bad_sha = CSRFixture(good, sha256sum=not_pem.sha256sum)

    bad_subject = CSRFixture(
        orgunit="Example test dept.",
        commonname="baz.example.com",
        pem=dedent("""\
            -----BEGIN CERTIFICATE REQUEST-----
            MIIBCDCBswIBADBOMRUwEwYDVQQKDAxFeGFtcGxlIGluYy4xGzAZBgNVBAsMEkV4
            YW1wbGUgdGVzdCBkZXB0LjEYMBYGA1UEAwwPYmF6LmV4YW1wbGUuY29tMFwwDQYJ
            KoZIhvcNAQEBBQADSwAwSAJBAOQ+LeU8pbEg5/XeD+HLnGah4A96qDL4HtNsTW0N
            //DbwGchtE8h9LSs3ePXWDVHak87Jx8A+wq7DRUd/LDVP6cCAwEAAaAAMA0GCSqG
            SIb3DQEBBQUAA0EAhriivXQYFbdc8QrnDjwVCX8ZGXiGKQEC66LceWDdvOy+mDu8
            gi+L6IgnptU8VmEowAPp0veIH1MWJrnGdp7M0g==
            -----END CERTIFICATE REQUEST-----
        """).encode("utf-8"),
        subject_components=(
            ("O", "Example inc."),
            ("OU", "Example test dept."),
            ("CN", "baz.example.com"),
        ),
        )

    large_body = CSRFixture(
        pem="".join(("foobar", "x" * views._MAXLEN)).encode("utf-8"),
        )



#! /usr/bin/env python
# vim: expandtab shiftwidth=4 softtabstop=4 tabstop=17 filetype=python :

# Make things as three-ish as possible (requires python >= 2.6)
from __future__ import (unicode_literals, print_function,
                        absolute_import, division)
# Namespace cleanup
del unicode_literals, print_function, absolute_import, division

#
# ----- End header -----
#

from pyramid.response import Response
from pyramid.httpexceptions import (
    HTTPLengthRequired,
    HTTPRequestEntityTooLarge,
    HTTPBadRequest,
    HTTPNotFound,
    HTTPForbidden,
    HTTPError
    )
from pyramid.view import view_config

from hashlib import sha256
from datetime import datetime

from .models import (
    CSR,
    AccessLog,
    SigningCert,
    )

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound

# Maximum length allowed for csr uploads.
# 2 kbyte should be enough for up to 4 kbit keys.
# XXX: This should probably be handled outside of app (i.e. by the
#      server), or at least be configurable.
_MAXLEN = 2 * 2**10


def raise_for_length(req, limit=_MAXLEN):
    # two possible error cases: no length specified, or length exceeds limit
    # raise appropriate exception if either applies
    length = req.content_length
    if length is None:
        raise HTTPLengthRequired
    if length > limit:
        raise HTTPRequestEntityTooLarge(
            "Max size: {0} kB".format(limit / 2**10)
            )


def raise_for_subject(components, required_prefix):
    if len(components) < len(required_prefix):
        raise ValueError("Too few subject components")
    result = [(x, y) for x, y in zip(components, required_prefix) if x != y]
    if result:
        given, required = zip(*result)
        raise ValueError("{0} do not match {1}".format(given, required))


# XXX: Is this the right way? Catch-class JSON converter of Exceptions
@view_config(context=HTTPError)
def HTTPErrorToJson(exc, request):
    exc.json_body = {
        "status": exc.code,
        "title": exc.title,
        "detail": exc.detail
    }
    exc.content_type = "application/problem+json"
    request.response = exc
    return request.response


@view_config(route_name="csr", request_method="POST", renderer="json")
def csr_add(request):
    # XXX: do length check in middleware? server?
    raise_for_length(request)
    sha256sum = sha256(request.body).hexdigest()
    if sha256sum != request.matchdict["sha256"]:
        raise HTTPBadRequest("hash mismatch ({0})".format(sha256sum))
    try:
        csr = CSR(sha256sum, request.body)
    except ValueError as err:
        raise HTTPBadRequest("crypto error: {0}".format(err))

    # Verify the parts of the subject we care about
    ca = SigningCert.from_files(request.registry.settings["ca.cert"])
    CA_PREFIX = ca.get_ca_prefix()
    try:
        raise_for_subject(csr.subject_components, CA_PREFIX)
    except ValueError as err:
        raise HTTPBadRequest("Bad subject: {0}".format(err))

    # XXX: store things in DB
    try:
        csr.save()
    except IntegrityError:
        # XXX: is this what we want here?
        raise HTTPBadRequest("duplicate request")
    # We've accepted the signing request, but there's been no signing yet
    request.response.status_int = 202
    # JSON-rendered data (client could calculate this itself, and often will)
    return csr


@view_config(route_name="cert", request_method="GET", renderer="json")
def cert_fetch(request):
    # XXX: JSON-renderer at the moment, to dump
    sha256sum = request.matchdict["sha256"]
    try:
        csr = CSR.by_sha256sum(sha256sum)
    except NoResultFound:
        raise HTTPNotFound
    # XXX: Exceptions? remote_addr or client_addr?
    AccessLog(csr, request.remote_addr).save()
    if csr.rejected:
        raise HTTPForbidden
    cert = csr.certificates.first()
    if cert:
        if datetime.utcnow() < cert.not_after:
            # XXX: appropriate content-type is ... ?
            return Response(cert.pem,
                            content_type="application/octet-stream",
                            charset="UTF-8")
    request.response.status_int = 202
    return csr


@view_config(route_name="ca", request_method="GET",
             renderer="string", http_cache=3600)
def ca_fetch(request):
    ca_file = request.registry.settings['ca.cert']
    ca = SigningCert.from_files(ca_file)
    return ca.pem.decode("utf8")


@view_config(route_name="cabundle", request_method="GET",
             renderer="string", http_cache=3600)
def ca_bundle_fetch(request):
    """Attempt to return a bunle of all our intermediates"""
    bundle = ca_fetch(request)
    return bundle



#! /usr/bin/env python
# vim: expandtab shiftwidth=4 softtabstop=4 tabstop=17 filetype=python :

# Make things as three-ish as possible (requires python >= 2.6)
from __future__ import (unicode_literals, print_function,
                        absolute_import, division)
# Namespace cleanup
del unicode_literals, print_function, absolute_import, division

#
# ----- End header -----
#

import unittest
import unittest.mock
import transaction
import datetime

from pyramid import testing
from pyramid.response import Response
from pyramid.httpexceptions import (
    HTTPLengthRequired,
    HTTPRequestEntityTooLarge,
    HTTPBadRequest,
    HTTPNotFound,
    )

from . import fixtures, ModelTestCase

from caramel.models import (
    init_session,
    DBSession,
    CSR,
    AccessLog,
    )
from caramel import views


def dummypost(fix, **args):
    req = testing.DummyRequest(**args)
    req.body = fix.pem
    req.content_length = len(req.body)
    req.matchdict["sha256"] = fix.sha256sum
    req.registry.settings['ca.cert'] = 'abc123.crt'
    return req


class TestCSRAdd(ModelTestCase):
    def setUp(self):
        super(TestCSRAdd, self).setUp()
        self.config = testing.setUp()
        # Store original
        self._from_files = views.SigningCert.from_files

        # mock instance that returns value
        _mocking = unittest.mock.Mock()
        _mocking.get_ca_prefix.return_value = fixtures.subject_prefix

        # Mock the object initializer
        views.SigningCert.from_files = unittest.mock.Mock()
        views.SigningCert.from_files.return_value = _mocking

    def tearDown(self):
        super(TestCSRAdd, self).tearDown()
        testing.tearDown()
        views.SigningCert.from_files = self._from_files

    def test_good(self):
        req = dummypost(fixtures.CSRData.good)
        csr = views.csr_add(req)
        self.assertSimilar(fixtures.CSRData.good, csr)

    def test_duplicate(self):
        req = dummypost(fixtures.CSRData.initial)
        with self.assertRaises(HTTPBadRequest):
            views.csr_add(req)

    def test_no_length(self):
        req = dummypost(fixtures.CSRData.good)
        req.content_length = None
        with self.assertRaises(HTTPLengthRequired):
            views.csr_add(req)

    def test_large(self):
        req = dummypost(fixtures.CSRData.large_body)
        with self.assertRaises(HTTPRequestEntityTooLarge):
            views.csr_add(req)

    def test_empty(self):
        req = dummypost(fixtures.CSRData.empty)
        with self.assertRaises(HTTPBadRequest):
            views.csr_add(req)

    def test_not_pem(self):
        req = dummypost(fixtures.CSRData.not_pem)
        with self.assertRaises(HTTPBadRequest):
            views.csr_add(req)

    def test_trailing_junk(self):
        req = dummypost(fixtures.CSRData.trailing_content)
        with self.assertRaises(HTTPBadRequest):
            views.csr_add(req)

    def test_leading_junk(self):
        req = dummypost(fixtures.CSRData.leading_content)
        with self.assertRaises(HTTPBadRequest):
            views.csr_add(req)

    def test_multi_pem(self):
        req = dummypost(fixtures.CSRData.multi_request)
        with self.assertRaises(HTTPBadRequest):
            views.csr_add(req)

    def test_bad_signature(self):
        req = dummypost(fixtures.CSRData.bad_signature)
        with self.assertRaises(HTTPBadRequest):
            views.csr_add(req)

    def test_bad_checksum(self):
        req = dummypost(fixtures.CSRData.bad_sha)
        with self.assertRaises(HTTPBadRequest):
            views.csr_add(req)

    def test_bad_subject(self):
        req = dummypost(fixtures.CSRData.bad_subject)
        with self.assertRaises(HTTPBadRequest):
            views.csr_add(req)


class TestCertFetch(ModelTestCase):
    def setUp(self):
        super(TestCertFetch, self).setUp()
        self.config = testing.setUp()
        self.req = testing.DummyRequest()
        self.req.remote_addr = "test"
        # TODO: ...

    def tearDown(self):
        super(TestCertFetch, self).tearDown()
        testing.tearDown()
        # TODO: ...

    def test_missing(self):
        self.req.matchdict["sha256"] = fixtures.CSRData.good.sha256sum
        accesses = len(AccessLog.all())
        with self.assertRaises(HTTPNotFound):
            views.cert_fetch(self.req)
        self.assertEqual(accesses, len(AccessLog.all()))

    def test_exists_valid(self):
        sha256sum = fixtures.CSRData.initial.sha256sum
        csr = CSR.by_sha256sum(sha256sum)
        now = datetime.datetime.utcnow()
        self.req.matchdict["sha256"] = sha256sum
        resp = views.cert_fetch(self.req)
        # Verify response contents
        self.assertIsInstance(resp, Response)
        self.assertEqual(resp.body, csr.certificates[0].pem)
        self.assertEqual(self.req.response.status_int, 200)
        # Verify there's a new AccessLog entry
        self.assertEqual(csr.accessed[0].addr, self.req.remote_addr)
        self.assertAlmostEqual(csr.accessed[0].when, now,
                               delta=datetime.timedelta(seconds=1))

    def test_exists_expired(self):
        csr = fixtures.CSRData.with_expired_cert()
        csr.save()
        now = datetime.datetime.utcnow()
        self.req.matchdict["sha256"] = csr.sha256sum
        resp = views.cert_fetch(self.req)
        # Verify response contents
        self.assertIs(resp, csr)
        self.assertEqual(self.req.response.status_int, 202)
        # Verify there's a new AccessLog entry
        self.assertEqual(csr.accessed[0].addr, self.req.remote_addr)
        self.assertAlmostEqual(csr.accessed[0].when, now,
                               delta=datetime.timedelta(seconds=1))

    def test_not_signed(self):
        csr = fixtures.CSRData.good()
        csr.save()
        now = datetime.datetime.utcnow()
        self.req.matchdict["sha256"] = csr.sha256sum
        resp = views.cert_fetch(self.req)
        # Verify response contents
        self.assertIs(resp, csr)
        self.assertEqual(self.req.response.status_int, 202)
        # Verify there's a new AccessLog entry
        self.assertEqual(csr.accessed[0].addr, self.req.remote_addr)
        self.assertAlmostEqual(csr.accessed[0].when, now,
                               delta=datetime.timedelta(seconds=1))
        pass


@unittest.skip("Example test case from scaffold, barely altered.")
class TestMyView(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp()
        from sqlalchemy import create_engine
        engine = create_engine("sqlite://")
        init_session(engine, create=True)
        from caramel.models import MyModel
        with transaction.manager:
            model = MyModel(name="one", value=55)
            model.save()

    def tearDown(self):
        DBSession.remove()
        testing.tearDown()

    def test_it(self):
        from caramel.views import my_view
        request = testing.DummyRequest()
        info = my_view(request)
        self.assertEqual(info["one"].name, "one")
        self.assertEqual(info["project"], "caramel")



#! /usr/bin/env python
# vim: expandtab shiftwidth=4 softtabstop=4 tabstop=17 filetype=python :

# Make things as three-ish as possible (requires python >= 2.6)
from __future__ import (unicode_literals, print_function,
                        absolute_import, division)
# Namespace cleanup
del unicode_literals, print_function, absolute_import, division

#
# ----- End header -----
#

import unittest
import transaction

from caramel.models import (
    init_session,
    DBSession,
    )

from . import fixtures

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest


class ModelTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(ModelTestCase, cls).setUpClass()
        # Clear existing session, if any.
        DBSession.remove()
        from sqlalchemy import create_engine
        engine = create_engine("sqlite://")
        init_session(engine, create=True)
        with transaction.manager:
            csr = fixtures.CSRData.initial()
            csr.save()

    def setUp(self):
        super(ModelTestCase, self).setUp()
        # Always run in a fresh session
        DBSession.remove()

    def assertSimilar(self, a, b, msg=None):
        if isinstance(b, fixtures.SimilarityComparable):
            a, b = b, a
        if isinstance(a, fixtures.SimilarityComparable):
            return self.assertTrue(a.match(b), msg)
        return self.assertEqual(a, b, msg)

    def assertSimilarSequence(self, seq1, seq2, msg=None):
        for a, b in zip_longest(seq1, seq2):
            self.assertSimilar(a, b)



#! /usr/bin/env python3
# vim: expandtab shiftwidth=4 softtabstop=4 tabstop=17 filetype=python :

import datetime
import argparse
import uuid
import os
import OpenSSL.crypto as _crypto
from pyramid.paster import bootstrap

VERSION = 0x2
CA_BITS = 4096
# Subject attribs, in order.
ATTRIBS_TO_KEEP = ("C", "ST", "L", "O", "OU", "CN")
CA_YEARS = 24  # Beware of unixtime ;)

CA_EXTENSIONS = [
    # Key usage for a CA cert.
    _crypto.X509Extension(b"basicConstraints", critical=True,
                          value=b"CA:true, pathlen:0"),
    # no cRLSign as we do not use CRLs in caramel.
    _crypto.X509Extension(b"keyUsage", critical=True, value=b"keyCertSign")
]


def _crypto_patch():
    """hijack _crypto internal lib and violate the default text encoding.
    https://github.com/pyca/pyopenssl/pull/115 has a pull&fix for it
    https://github.com/pyca/pyopenssl/issues/129 is an open issue
    about it."""
    _crypto._lib.ASN1_STRING_set_default_mask_asc(b'utf8only')


_crypto_patch()


# Hack hack. :-)
def CA_LIFE():
    d = datetime.date.today()
    t = datetime.date(d.year+CA_YEARS, d.month, d.day)
    return int((t-d).total_seconds())


# adapted from models.py
def components(subject):
    comps = subject.get_components()
    return dict((n.decode("utf8"), v.decode("utf8")) for n, v in comps)


def matching_template(x509, cacert):
    """ Takes a subject as a dict, and returns if all required fields
    match. Otherwise raises exception"""

    def later_check(subject):
        """Check that the last two fields in subject are OU, CN"""
        pair = subject[-1]
        if pair[0].decode("utf8") != "CN":
            raise ValueError("CN needs to be last in subject")

        pair = subject[-2]
        if pair[0].decode("utf8") != "OU":
            raise ValueError("OU needs to be second to last")

    casubject = cacert.get_subject().get_components()
    subject = x509.get_subject().get_components()
    later_check(casubject)
    later_check(subject)

    casubject = casubject[:-2]
    subject = subject[:-2]

    for (ca, sub) in zip(casubject, subject):
        if ca != sub:
            raise ValueError("Subject needs to match CA cert:"
                             "{}".format(casubject))


def sign_req(req, cacert, cakey):
    # Validate Subject contents. Not necessary for CA gen, but kept anyhow
    matching_template(req, cacert)

    # Validate signature
    req.verify(req.get_pubkey())
    request_subject = components(req.get_subject())

    cert = _crypto.X509()
    subject = cert.get_subject()
    cert.set_serial_number(int(uuid.uuid1()))
    cert.set_version(VERSION)

    for attrib in ATTRIBS_TO_KEEP:
        if request_subject.get(attrib):
            setattr(subject, attrib, request_subject[attrib])

    issuer_subject = cert.get_subject()
    cert.set_issuer(issuer_subject)
    cert.set_pubkey(req.get_pubkey())

    # Validity times
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(CA_LIFE())

    cert.add_extensions(CA_EXTENSIONS)

    cacert = cert

    extension = _crypto.X509Extension(b"subjectKeyIdentifier",
                                      critical=False,
                                      value=b"hash",
                                      subject=cert)
    cert.add_extensions([extension])

    # We need subjectKeyIdentifier to be added before we can add
    # authorityKeyIdentifier.
    extension = _crypto.X509Extension(b"authorityKeyIdentifier",
                                      critical=False,
                                      value=b"issuer:always,keyid:always",
                                      issuer=cacert)
    cert.add_extensions([extension])

    cert.sign(cakey, "sha512")
    return cert


def create_ca_req(subject):
    key = _crypto.PKey()
    key.generate_key(_crypto.TYPE_RSA, CA_BITS)

    req = _crypto.X509Req()
    req.set_version(VERSION)
    req.set_pubkey(key)

    x509subject = req.get_subject()
    for k, v in subject:
        setattr(x509subject, k, v)

    req.add_extensions(CA_EXTENSIONS)

    req.sign(key, "sha512")
    return key, req


def create_ca(subject):
    key, req = create_ca_req(subject)
    cert = sign_req(req, req, key)
    return key, req, cert


def write_files(key, keyname, cert, certname):

    def writefile(data, name):
        with open(name, 'w') as f:
            stream = data.decode('utf8')
            f.write(stream)

    _key = _crypto.dump_privatekey(_crypto.FILETYPE_PEM, key)
    writefile(_key, keyname)

    _cert = _crypto.dump_certificate(_crypto.FILETYPE_PEM, cert)
    writefile(_cert, certname)


def cmdline():
    parser = argparse.ArgumentParser()
    parser.add_argument("inifile", help="Needed to find CA filename")
    args = parser.parse_args()
    return args


def build_ca(keyname, certname):
    print("Enter CA settings, leave blank to not include")
    subject = {}
    subject["C"] = input("C [countryName (Code, 2 letters)]: ").upper()
    if subject["C"] and len(subject["C"]) != 2:
        raise ValueError("Country codes are two letters")

    subject["ST"] = input("ST [stateOrProvinceName]: ")[:20]
    subject["L"] = input("L [localityName]: ")
    subject["O"] = input("O [Organization]: ")
    subject["OU"] = input("OU [organizationalUnitName]: ") or "Caramel"
    subject["CN"] = "Caramel Signing Certificate"
    print("CN will be '{}'".format(subject["CN"]))

    template = []
    for field in ATTRIBS_TO_KEEP:
        if field in subject and subject[field]:
            template.append((field, subject[field]))
    template = tuple(template)

    key, req, cert = create_ca(template)
    write_files(key=key, keyname=keyname,
                cert=cert, certname=certname)


def main():
    args = cmdline()
    env = bootstrap(args.inifile)
    settings, closer = env["registry"].settings, env["closer"]

    ca_cert = settings.get("ca.cert")
    ca_key = settings.get("ca.key")
    if not ca_cert:
        print("Missing 'ca.cert' in ini file")
        closer()
        exit()

    if not ca_key:
        print("Missing 'ca.cert' in ini file")
        closer()
        exit()

    for f in ca_cert, ca_key:
        if os.path.exists(f):
            print("File already exists: {}. Refusing to corrupt.".format(f))
            closer()
            exit()
        else:
            dname = os.path.dirname(f)
            os.makedirs(dname, exist_ok=True)

    print("Will write key to {}".format(ca_key))
    print("Will write cert to {}".format(ca_cert))

    build_ca(keyname=ca_key, certname=ca_cert)



#!/bin/env python3

"""The caramel auto-signer daemon.
This isn't necessary a _good_ idea, but it's part of examples of how
you can use caramel for application signups.

To work, you need to embed the root CA in your application (effectively pinning
it).
The client then generates a key (if none exists) and a CSR (based on the stored
CA). It pushes this to the caramel server, and goes into a spin-loop, waiting
for the key to get signed.

The autosigner then automatically signs the CSR.
The client gets it's certificate, and then connects to a second service (not
quite implemented here)  where the user adds contact information
(email/password) which ties an identity to the private key.

This can then be used to link multiple keys/applications to the same user, in a
safe & secure manner.

Autosigneer is implemented as it's own service, to make it obvious that you
shouldn't have your private key accessible by the web-application.
it."""

import argparse
import transaction
import sys
import logging
import datetime
import time
import uuid
import concurrent.futures

from pyramid.paster import bootstrap
from sqlalchemy import create_engine

import caramel.models as models

logger = logging.getLogger(__name__)


def csr_sign(csr, ca, delta):
    """Signs a CSR and saves it in a transaction.
    Transaction so we won't have racing with the database.
    Also validates that it's a UUID for commonname."""

    # Could have been by us, or before
    if csr.rejected:
        return

    try:
        uuid.UUID(csr.commonname)
    except ValueError:
        # not a valid uuid. Just ignore
        return

    with transaction.manager:
        cert = models.Certificate.sign(csr, ca, delta)
        cert.save()
    return


def mainloop(delay, ca, delta):
    """Concurrent-enabled mainloop.
    Spins forever and signs all certificates that come in"""
    with concurrent.futures.ThreadPoolExecutor(max_workers=16) as executor:
        while True:
            csrs = models.CSR.unsigned()
            futures = [executor.submit(csr_sign, csr, ca, delta)
                       for csr in csrs]

            time.sleep(delay)
            for future in concurrent.futures.as_completed(futures):
                try:
                    future.result()
                except Exception:
                    logger.exception("Future failed")


def cmdline():
    """Basically just parsing the arguments and returning them"""
    parser = argparse.ArgumentParser()
    parser.add_argument("inifile")
    parser.add_argument("--delay", help="How long to sleep. (ms)")
    parser.add_argument("--valid",
                        help="How many hours the certificate is valid for")
    args = parser.parse_args()
    return args


def error_out(message, closer):
    """Just log a message, and perform cleanup"""
    logger.error(message)
    closer()
    sys.exit(1)


def main():
    """Main, as called from the script instance by pyramid"""
    logging.basicConfig()
    logger.setLevel(logging.DEBUG)
    args = cmdline()
    env = bootstrap(args.inifile)
    settings, closer = env['registry'].settings, env['closer']
    engine = create_engine(settings['sqlalchemy.url'])
    models.init_session(engine)
    delay = int(settings.get('delay', 500)) / 1000
    valid = int(settings.get('valid', 3))
    delta = datetime.timedelta(days=0, hours=valid)
    del valid

    try:
        certname = settings["ca.cert"]
        keyname = settings["ca.key"]
    except KeyError:
        error_out("config file needs ca.cert and ca.key properly set", closer)
    ca = models.SigningCert.from_files(certname, keyname)
    mainloop(delay, ca, delta)


if __name__ == "__main__":
    logging.basicConfig()



#! /usr/bin/env python
# vim: expandtab shiftwidth=4 softtabstop=4 tabstop=17 filetype=python :

# Make things as three-ish as possible (requires python >= 2.6)
from __future__ import (unicode_literals, print_function,
                        absolute_import, division)
# Namespace cleanup
del unicode_literals, print_function, absolute_import, division

#
# ----- End header -----
#

import unittest

from caramel.models import (
    CSR,
    SigningCert,
)

from . import fixtures, ModelTestCase

from operator import attrgetter
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound


class TestCSR(ModelTestCase):
    def test_good(self):
        csrs = [fixtures.CSRData.initial]
        self.assertSimilarSequence(CSR.all(), csrs)
        csrs.append(fixtures.CSRData.good)
        csrs[-1]().save()
        sortkey = attrgetter("sha256sum")
        csrs.sort(key=sortkey)
        self.assertSimilarSequence(sorted(CSR.all(), key=sortkey), csrs)
        self.assertTrue(csrs[1].match(CSR.by_sha256sum(csrs[1].sha256sum)))

    def test_not_found(self):
        with self.assertRaises(NoResultFound):
            CSR.by_sha256sum(fixtures.CSRData.good.sha256sum)

    def test_not_unique(self):
        with self.assertRaises(IntegrityError):
            fixtures.CSRData.initial().save()

    def test_bad_signature(self):
        with self.assertRaises(ValueError):
            fixtures.CSRData.bad_signature().save()

    def test_truncated(self):
        with self.assertRaises(ValueError):
            fixtures.CSRData.truncated().save()

    def test_trailing_content(self):
        with self.assertRaises(ValueError):
            fixtures.CSRData.trailing_content().save()

    def test_multi_request(self):
        with self.assertRaises(ValueError):
            fixtures.CSRData.multi_request().save()

    def test_not_pem(self):
        with self.assertRaises(ValueError):
            fixtures.CSRData.not_pem().save()

    def test_empty(self):
        with self.assertRaises(ValueError):
            fixtures.CSRData.empty().save()


class TestGetCAPrefix(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_blank_file(self):
        import OpenSSL
        with self.assertRaises(OpenSSL.crypto.Error):
            SigningCert("")

    def test_valid_cert(self):
        ca = SigningCert(fixtures.CertificateData.ca_cert.pem)
        ca.get_ca_prefix()

    def test_outdated_cert_should_work(self):
        ca = SigningCert(fixtures.CertificateData.expired.pem)
        ca.get_ca_prefix()

    def test_empty_returns_from_empty_subject(self):
        ca = SigningCert(fixtures.CertificateData.initial.pem)
        result = ca.get_ca_prefix()
        self.assertEqual((), result)

    def test_empty_returns_from_empty_selector(self):
        ca = SigningCert(fixtures.CertificateData.ca_cert.pem)
        result = ca.get_ca_prefix(())
        self.assertEqual((), result)

    def test_valid_returns_from_default_subject(self):
        ca = SigningCert(fixtures.CertificateData.ca_cert.pem)
        r = ca.get_ca_prefix()
        self.assertEqual(fixtures.CertificateData.ca_cert.common_subject, r)

    def test_only_CN_returns_from_CN_selector(self):
        CN_TUPLE = (('CN', 'Caramel Signing Certificate'),)
        ca = SigningCert(fixtures.CertificateData.ca_cert.pem)
        result = ca.get_ca_prefix((b'CN', ))
        self.assertEqual(CN_TUPLE, result)

    def test_only_wanted_returns_from_selector(self):
        SELECTED = (('ST', 'Östergötland'),
                    ('L', 'Norrköping'),
                    ('OU', 'Muppar Teknik'))
        SELECTOR = (b'ST', b'L', b'OU')
        ca = SigningCert(fixtures.CertificateData.ca_cert.pem)
        result = ca.get_ca_prefix(SELECTOR)
        self.assertEqual(SELECTED, result)



#!/bin/env python3
# vim: expandtab shiftwidth=4 softtabstop=4 tabstop=17 filetype=python :

import argparse

from pyramid.paster import bootstrap
from pyramid.settings import asbool
from sqlalchemy import create_engine
from dateutil.relativedelta import relativedelta
import caramel.models as models
import transaction
import datetime
import sys
import concurrent.futures


def cmdline():
    parser = argparse.ArgumentParser()
    parser.add_argument("inifile")
    parser.add_argument(
        "--long",
        help="Generate a long lived cert(1 year)",
        action="store_true",
    )

    parser.add_argument(
        "--list",
        help="List active requests, do nothing else",
        action="store_true",
    )

    exclusives = parser.add_mutually_exclusive_group()
    exclusives.add_argument(
        "--sign", metavar="id", type=int, help="Sign the CSR with this id"
    )
    exclusives.add_argument(
        "--reject",
        metavar="id",
        type=int,
        help="Reject the CSR with this id",
    )

    cleanout = parser.add_mutually_exclusive_group()
    cleanout.add_argument(
        "--clean",
        metavar="id",
        type=int,
        help="Remove all older certificates for this CSR",
    )
    cleanout.add_argument(
        "--wipe",
        metavar="id",
        type=int,
        help="Wipe all certificates for this CSR",
    )

    bulk = parser.add_mutually_exclusive_group()
    bulk.add_argument(
        "--refresh",
        help="Sign all certificates that have a valid current signature.",
        action="store_true",
    )

    bulk.add_argument(
        "--cleanall",
        help="Clean all older certificates.",
        action="store_true",
    )

    args = parser.parse_args()

    # Didn't find a way to do this with argparse, but I didn't look too hard.
    return args


def error_out(message):
    print(message)
    sys.exit(1)


def print_list():
    requests = models.CSR.valid()

    def unsigned_last(csr):
        return (not csr.certificates, csr.id)

    requests.sort(key=unsigned_last)

    for csr in requests:
        cert = csr.certificates.first()
        not_after = "----------"
        if cert:
            not_after = str(cert.not_after)
        output = " ".join(
            (str(csr.id), csr.commonname, csr.sha256sum, not_after)
        )
        # TODO: Add lifetime of latest (fetched?) cert for the key.
        print(output)


def calc_lifetime(lifetime=relativedelta(hours=24)):
    now = datetime.datetime.utcnow()
    future = now + lifetime
    return future - now


def csr_wipe(csr_id):
    with transaction.manager:
        CSR = models.CSR.query().get(csr_id)
        if not CSR:
            error_out("ID not found")
        CSR.certificates = []
        CSR.save()


def csr_clean(csr_id):
    with transaction.manager:
        CSR = models.CSR.query().get(csr_id)
        if not CSR:
            error_out("ID not found")
        certs = [CSR.certificates.first()]
        CSR.certificates = certs
        CSR.save()


def clean_all():
    csrlist = models.CSR.refreshable()
    for csr in csrlist:
        csr_clean(csr.id)


def csr_reject(csr_id):
    with transaction.manager:
        CSR = models.CSR.query().get(csr_id)
        if not CSR:
            error_out("ID not found")

        CSR.rejected = True
        CSR.save()


def csr_sign(csr_id, ca, timedelta, backdate):
    with transaction.manager:
        CSR = models.CSR.query().get(csr_id)
        if not CSR:
            error_out("ID not found")
        if CSR.rejected:
            error_out("Refusing to sign rejected ID")

        cert = CSR.certificates.first()
        if cert:
            today = datetime.datetime.utcnow()
            cur_lifetime = cert.not_after - cert.not_before
            # Cert hasn't expired, and currently has longer lifetime
            if (cert.not_after > today) and (cur_lifetime > timedelta):
                msg = (
                    "Currently has a valid certificate with {} lifetime, "
                    "new certificate would have {} lifetime. \n"
                    "Clean out existing certificates before shortening "
                    "lifetime.\n"
                    "The old certificate is still out there."
                )
                error_out(msg.format(cur_lifetime, timedelta))

        cert = models.Certificate.sign(CSR, ca, timedelta, backdate)
        cert.save()


def refresh(csr, ca, lifetime_short, lifetime_long, backdate):
    last = csr.certificates.first()
    old_lifetime = last.not_after - last.not_before
    # XXX: In a backdated cert, this is almost always true.
    if old_lifetime >= lifetime_long:
        cert = models.Certificate.sign(csr, ca, lifetime_long, backdate)
    else:
        # Never backdate short-lived certs
        cert = models.Certificate.sign(csr, ca, lifetime_short, False)
    with transaction.manager:
        cert.save()


def csr_resign(ca, lifetime_short, lifetime_long, backdate):
    with concurrent.futures.ThreadPoolExecutor(max_workers=16) as executor:
        try:
            csrlist = models.CSR.refreshable()
        except Exception:
            error_out("Not found or some other error")
        futures = (
            executor.submit(
                refresh, csr, ca, lifetime_short, lifetime_long, backdate
            )
            for csr in csrlist
        )
        for future in concurrent.futures.as_completed(futures):
            try:
                future.result()
            except Exception:
                print("Future failed")


def main():
    args = cmdline()
    env = bootstrap(args.inifile)
    settings, closer = env["registry"].settings, env["closer"]
    engine = create_engine(settings["sqlalchemy.url"])
    models.init_session(engine)
    settings_backdate = asbool(settings.get("backdate", False))

    _short = int(settings.get("lifetime.short", 48))
    _long = int(settings.get("lifetime.long", 7 * 24))
    life_short = calc_lifetime(relativedelta(hours=_short))
    life_long = calc_lifetime(relativedelta(hours=_long))
    del _short, _long

    try:
        certname = settings["ca.cert"]
        keyname = settings["ca.key"]
    except KeyError:
        error_out("config file needs ca.cert and ca.key properly set")
    ca = models.SigningCert.from_files(certname, keyname)

    if life_short > life_long:
        error_out(
            "Short lived certs ({0}) shouldn't last longer "
            "than long lived certs ({1})".format(life_short, life_long)
        )
    if args.list:
        print_list()
        closer()
        sys.exit(0)

    if args.reject:
        csr_reject(args.reject)

    if args.wipe:
        error_out("Not implemented yet")

    if args.clean:
        error_out("Not implemented yet")

    if args.cleanall:
        clean_all()

    if args.sign:
        if args.long:
            csr_sign(args.sign, ca, life_long, settings_backdate)
        else:
            # Never backdate short lived certs
            csr_sign(args.sign, ca, life_short, False)

    if args.refresh:
        csr_resign(ca, life_short, life_long, settings_backdate)
